# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
from enum import Enum


# status for user is super
# 是否为超级管理员的枚举
class UserSuper(Enum):
    COMMON = 1
    SUPER = 2


# : status for user is active
# : 当前用户是否为激活状态的枚举
class UserActive(Enum):
    ACTIVE = 1
    NOT_ACTIVE = 2



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

from collections import defaultdict


class Config(defaultdict):
    def add_plugin_config(self, plugin_name, obj):
        if type(obj) is dict:
            if self.get(plugin_name, None) is None:
                self[plugin_name] = {}
            for k, v in obj.items():
                self[plugin_name][k] = v

    def add_plugin_config_item(self, plugin_name, key, value):
        if self.get(plugin_name, None) is None:
            self[plugin_name] = {}
        self[plugin_name][key] = value

    def get_plugin_config(self, plugin_name, default=None):
        return self.get(plugin_name, default)

    def get_plugin_config_item(self, plugin_name, key, default=None):
        plugin_conf = self.get(plugin_name)
        if plugin_conf is None:
            return default
        return plugin_conf.get(key, default)

    def get_config(self, key: str, default=None):
        """ plugin_name.key """
        if '.' not in key:
            return self.get(key, default)
        index = key.rindex('.')
        plugin_name = key[:index]
        plugin_key = key[index + 1:]
        plugin_conf = self.get(plugin_name)
        if plugin_conf is None:
            return default
        return plugin_conf.get(plugin_key, default)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

from collections import namedtuple
from datetime import datetime, date

from flask import Flask, current_app, request, Blueprint
from flask.json import JSONEncoder as _JSONEncoder
from werkzeug.exceptions import HTTPException
from werkzeug.local import LocalProxy

from .db import db
from .jwt import jwt
from .exception import APIException, UnknownException
from .interface import UserInterface, GroupInterface, AuthInterface, LogInterface, EventInterface
from .exception import NotFound, AuthFailed
from .config import Config

__version__ = 0.1

# 路由函数的权限和模块信息(meta信息)
Meta = namedtuple('meta', ['auth', 'module'])

#       -> endpoint -> func
# auth                      -> module
#       -> endpoint -> func

# 记录路由函数的权限和模块信息
route_meta_infos = {}


gigi_config = Config()


def route_meta(auth, module='common', mount=True):
    """
    记录路由函数的信息
    记录路由函数访问的推送信息模板
    注：只有使用了 route_meta 装饰器的函数才会被记录到权限管理的map中
    :param auth: 权限
    :param module: 所属模块
    :param mount: 是否挂在到权限中（一些视图函数需要说明，或暂时决定不挂在到权限中，则设置为False）
    :return:
    """

    def wrapper(func):
        if mount:
            name = func.__name__
            exit = route_meta_infos.get(name, None)
            if exit:
                raise Exception("func's name cant't be repeat")
            else:
                route_meta_infos.setdefault(name, Meta(auth, module))
        return func

    return wrapper


def find_user(**kwargs):
    return manager.find_user(**kwargs)


def find_group(**kwargs):
    return manager.find_group(**kwargs)


def get_ep_infos():
    """ 返回权限管理中的所有视图函数的信息，包含它所属module """
    infos = {}
    for ep, meta in manager.ep_meta.items():
        mod = infos.get(meta.module, None)
        if mod:
            sub = mod.get(meta.auth, None)
            if sub:
                sub.append(ep)
            else:
                mod[meta.auth] = [ep]
        else:
            infos.setdefault(meta.module, {meta.auth: [ep]})

    return infos


def find_info_by_ep(ep):
    """ 通过请求的endpoint寻找路由函数的meta信息"""
    return manager.ep_meta.get(ep)


def is_user_allowed(group_id):
    """查看当前user有无权限访问该路由函数"""
    ep = request.endpoint
    # 根据 endpoint 查找 authority
    meta = manager.ep_meta.get(ep)
    return manager.verity_user_in_group(group_id, meta.auth, meta.module)


def find_auth_module(auth):
    """ 通过权限寻找meta信息"""
    for _, meta in manager.ep_meta.items():
        if meta.auth == auth:
            return meta
    return None


class Jian(object):

    def __init__(self,
                 app: Flask = None,  # flask app , default None
                 group_model=None,  # group model, default None
                 user_model=None,  # user model, default None
                 auth_model=None,  # authority model, default None
                 create_all=False,  # 是否创建所有数据库表, default false
                 mount=True,  # 是否挂载默认的蓝图, default True
                 handle=True,  # 是否使用全局异常处理 , default True
                 json_encoder=True,  # 是否使用自定义的json_encoder , default True
                 ):
        self.app = app
        self.manager = None
        if app is not None:
            self.init_app(app, group_model, user_model, auth_model, create_all, mount, handle, json_encoder)

    def init_app(self,
                 app: Flask,
                 group_model=None,
                 user_model=None,
                 auth_model=None,
                 create_all=False,
                 mount=True,
                 handle=True,
                 json_encoder=True,
                 ):
        # default config
        app.config.setdefault('PLUGIN_PATH', {})
        # 默认蓝图的前缀
        app.config.setdefault('BP_URL_PREFIX', '')
        json_encoder and self._enable_json_encoder(app)
        self.app = app
        # 初始化 manager
        self.manager = Manager(app.config.get('PLUGIN_PATH'),
                               group_model,
                               user_model,
                               auth_model)
        self.app.extensions['manager'] = self.manager
        db.init_app(app)
        create_all and self._enable_create_all(app)
        jwt.init_app(app)
        mount and self.mount(app)
        handle and self.handle_error(app)

    def mount(self, app):
        # 加载默认插件路由
        bp = Blueprint('plugin', __name__)
        # 加载插件的路由
        for plugin in self.manager.plugins.values():
            for controller in plugin.controllers.values():
                controller.register(bp)
        app.register_blueprint(bp, url_prefix=app.config.get('BP_URL_PREFIX'))
        for ep, func in app.view_functions.items():
            info = route_meta_infos.get(func.__name__, None)
            if info:
                self.manager.ep_meta.setdefault(ep, info)

    def handle_error(self, app):
        @app.errorhandler(Exception)
        def handler(e):
            if isinstance(e, APIException):
                return e
            if isinstance(e, HTTPException):
                code = e.code
                msg = e.description
                error_code = 20000
                return APIException(msg, code, error_code)
            else:
                if not app.config['DEBUG']:
                    return UnknownException()
                else:
                    raise e

    def _enable_json_encoder(self, app):
        app.json_encoder = JSONEncoder

    def _enable_create_all(self, app):
        with app.app_context():
            db.create_all()


class Manager(object):
    """ manager for Jian """ 

    # 路由函数的meta信息的容器
    ep_meta = {}

    def __init__(self, plugin_path, group_model=None, user_model=None, auth_model=None):
        if not group_model:
            self.group_model = Group
        else:
            self.group_model = group_model

        if not user_model:
            self.user_model = User
        else:
            self.user_model = user_model

        if not auth_model:
            self.auth_model = Auth
        else:
            self.auth_model = auth_model

        from .loader import Loader
        self.loader: Loader = Loader(plugin_path)

    def find_user(self, **kwargs):
        return self.user_model.query.filter_by(**kwargs).first()

    def verify_user(self, nickname, password):
        return self.user_model.verify(nickname, password)

    def find_group(self, **kwargs):
        return self.group_model.query.filter_by(**kwargs).first()

    def verity_user_in_group(self, group_id, auth, module):
        return self.auth_model.query.filter_by(group_id=group_id, auth=auth, module=module).first()

    @property
    def plugins(self):
        return self.loader.plugins

    def get_plugin(self, name):
        return self.loader.plugins.get(name)

    def get_model(self, name):
        # attention!!! if models have the same name,will return the first one
        # 注意！！！ 如果容器内有相同的model，则默认返回第一个
        for plugin in self.plugins.values():
            return plugin.models.get(name)

    def get_service(self, name):
        # attention!!! if services have the same name,will return the first one
        # 注意！！！ 如果容器内有相同的service，则默认返回第一个
        for plugin in self.plugins.values():
            return plugin.services.get(name)


# a proxy for manager instance
# attention, only used when context in  stack

# 获得manager实例
# 注意，仅仅在flask的上下文栈中才可获得
manager: Manager = LocalProxy(lambda: get_manager())


def get_manager():
    _manager = current_app.extensions['manager']
    if _manager:
        return _manager
    else:
        app = current_app._get_current_object()
        with app.app_context():
            return app.extensions['manager']




class User(UserInterface, db.Model):

    @classmethod
    def verify(cls, nickname, password):
        user = cls.query.filter_by(nickname=nickname).first()
        if user is None:
            raise NotFound(msg='用户不存在')
        if not user.check_password(password):
            raise AuthFailed(msg='密码错误，请输入正确密码')
        return user

    def reset_password(self, new_password):
        #: attention,remember to commit
        #: 注意，修改密码后记得提交至数据库
        self.password = new_password

    def change_password(self, old_password, new_password):
        #: attention,remember to commit
        #: 注意，修改密码后记得提交至数据库
        if self.check_password(old_password):
            self.password = new_password
            return True
        return False


class Group(GroupInterface):
    pass


class Auth(AuthInterface):
    pass


# log model
class Log(LogInterface):
    @staticmethod
    def create_log(**kwargs):
        log = Log()
        for key in kwargs.keys():
            if hasattr(log, key):
                setattr(log, key, kwargs[key])
        db.session.add(log)
        if kwargs.get('commit') is True:
            db.session.commit()
        return log


# event model
class Event(EventInterface):
    pass


class JSONEncoder(_JSONEncoder):
    def default(self, o):
        if hasattr(o, 'keys') and hasattr(o, '__getitem__'):
            return dict(o)
        if isinstance(o, datetime):
            return o.strftime('%Y-%m-%dT%H:%M:%SZ')
        if isinstance(o, date):
            return o.strftime('%Y-%m-%d')
        return JSONEncoder.default(self, o)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
import time
import re
import errno
import random
import types
from importlib import import_module
import os
import importlib.util
from flask import request, current_app

from .exception import ParameterException


def get_timestamp(fmt='%Y-%m-%d %H:%M:%S'):
    return time.strftime(fmt, time.localtime(time.time()))


def get_pyfile(path, module_name, silent=False):
    """
    get all properties of a pyfile
    获得一个.py文件的所有属性
    :param path: path of pytfile
    :param module_name: name
    :param silent: show the error or not
    :return: all properties of a pyfile
    """
    d = types.ModuleType(module_name)
    d.__file__ = path
    try:
        with open(path, mode='rb') as config_file:
            exec(compile(config_file.read(), path, 'exec'), d.__dict__)
    except IOError as e:
        if silent and e.errno in (
                errno.ENOENT, errno.EISDIR, errno.ENOTDIR
        ):
            return False
        e.strerror = 'Unable to load configuration file (%s)' % e.strerror
        raise
    return d.__dict__


def load_object(path):
    """
    获得一个模块中的某个属性
    :param path: module path
    :return: the obj of module which you want get.
    """
    try:
        dot = path.rindex('.')
    except ValueError:
        raise ValueError("Error loading object '%s': not a full path" % path)

    module, name = path[:dot], path[dot + 1:]
    mod = import_module(module)

    try:
        obj = getattr(mod, name)
    except AttributeError:
        raise NameError("Module '%s' doesn't define any object named '%s'" % (module, name))

    return obj


def import_module_abs(name, path):
    """
    绝对路径导入模块
    :param name: name of module
    :param path: absolute path of module
    :return: the module
    """
    spec = importlib.util.spec_from_file_location(name, path)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)


def get_pwd():
    """
    :return: absolute current work path
    """
    return os.path.abspath(os.getcwd())


def paginate():
    count = int(request.args.get('count', current_app.config.get('COUNT_DEFAULT') if current_app.config.get(
        'COUNT_DEFAULT') else 5))
    start = int(request.args.get('page', current_app.config.get('PAGE_DEFAULT') if current_app.config.get(
        'PAGE_DEFAULT') else 0))
    count = 15 if count >= 15 else count
    start = start * count
    if start < 0 or count < 0:
        raise ParameterException()
    return start, count


def camel2line(camel: str):
    p = re.compile(r'([a-z]|\d)([A-Z])')
    line = re.sub(p, r'\1_\2', camel).lower()
    return line


def get_random_str(length):
    seed = "1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    sa = []
    for i in range(length):
        sa.append(random.choice(seed))
    salt = ''.join(sa)
    return salt



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
from importlib import import_module

from .redprint import Redprint
from .db import db
from .plugin import Plugin
from .core import gigi_config


class Loader(object):
    plugin_path: dict = None

    def __init__(self, plugin_path):
        self.plugins = {}
        assert type(plugin_path) is dict, 'plugin_path must be a dict'
        self.plugin_path = plugin_path
        self.load_plugins_config()
        self.load_plugins()

    def load_plugins(self):
        for name, conf in self.plugin_path.items():
            enable = conf.get('enable', None)
            if enable:
                path = conf.get('path')
                # load plugin
                path and self._load_plugin(f'{path}.app.__init__', name)

    def load_plugins_config(self):
        for name, conf in self.plugin_path.items():
            path = conf.get('path', None)
            # load config
            self._load_config(f'{path}.config', name, conf)

    def _load_plugin(self, path, name):
        mod = import_module(path)
        plugin = Plugin(name=name)
        dic = mod.__dict__
        for key in dic.keys():
            if not key.startswith('_'):
                attr = dic[key]
                if isinstance(attr, Redprint):
                    plugin.add_controller(attr.name, attr)
                elif issubclass(attr, db.Model):
                    plugin.add_model(attr.__name__, attr)
                # 暂时废弃加载service，用处不大
                # elif issubclass(attr, ServiceInterface):
                #     plugin.add_service(attr.__name__, attr)
        self.plugins[plugin.name] = plugin

    def _load_config(self, config_path, name, conf):
        default_conf = {**conf} if conf else {}
        try:
            if config_path:
                mod = import_module(config_path)
                dic = mod.__dict__
                for key in dic.keys():
                    if not key.startswith('_'):
                        default_conf[key] = dic[key]
        except ModuleNotFoundError as e:
            pass
        gigi_config.add_plugin_config(plugin_name=name, obj=default_conf)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
from flask import json, request
from werkzeug.exceptions import HTTPException
from werkzeug._compat import text_type


class APIException(HTTPException):
    code = 500
    msg = '抱歉，服务器未知错误'
    error_code = 999

    def __init__(self, msg=None, code=None, error_code=None,
                 headers=None):
        if code:
            self.code = code
        if error_code:
            self.error_code = error_code
        if msg:
            self.msg = msg
        if headers is not None:
            headers_merged = headers.copy()
            headers_merged.update(self.headers)
            self.headers = headers_merged

        super(APIException, self).__init__(msg, None)

    def get_body(self, environ=None):
        body = dict(
            msg=self.msg,
            error_code=self.error_code,
            request=request.method + '  ' + self.get_url_no_param()
        )
        text = json.dumps(body)
        return text_type(text)

    @staticmethod
    def get_url_no_param():
        full_path = str(request.full_path)
        main_path = full_path.split('?')
        return main_path[0]

    def get_headers(self, environ=None):
        return [('Content-Type', 'application/json')]


class Success(APIException):
    code = 201
    msg = '成功'
    error_code = 0


class Failed(APIException):
    code = 400
    msg = '失败'
    error_code = 9999


class AuthFailed(APIException):
    code = 401
    msg = '认证失败'
    error_code = 10000


class NotFound(APIException):
    code = 404
    msg = '资源不存在'
    error_code = 10020


class ParameterException(APIException):
    code = 400
    msg = '参数错误'
    error_code = 10030


class InvalidTokenException(APIException):
    code = 401
    msg = '令牌失效'
    error_code = 10040


class ExpiredTokenException(APIException):
    code = 422
    msg = '令牌过期'
    error_code = 10050


class UnknownException(APIException):
    code = 400
    msg = '服务器未知错误'
    error_code = 999


class RepeatException(APIException):
    code = 400
    msg = '字段重复'
    error_code = 10060


class Forbidden(APIException):
    code = 401
    msg = '不可操作'
    error_code = 10070



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""


class Redprint:
    def __init__(self, name, with_prefix=True):
        self.name = name
        self.with_prefix = with_prefix
        self.mound = []

    def route(self, rule, **options):
        def decorator(f):
            self.mound.append((f, rule, options))
            return f

        return decorator

    def register(self, bp, url_prefix=None):
        if url_prefix is None and self.with_prefix:
            url_prefix = '/' + self.name
        else:
            url_prefix = ''
        for f, rule, options in self.mound:
            endpoint = self.name + '+' + options.pop("endpoint", f.__name__)
            if rule:
                url = url_prefix + rule
                bp.add_url_rule(url, endpoint, f, **options)
            else:
                bp.add_url_rule(url_prefix, endpoint, f, **options)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

from functools import wraps

from flask_jwt_extended import JWTManager, verify_jwt_in_request, get_current_user, create_access_token, \
    create_refresh_token

from .exception import AuthFailed, InvalidTokenException, ExpiredTokenException, NotFound

jwt = JWTManager()


def admin_required(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        verify_jwt_in_request()
        current_user = get_current_user()
        if not current_user.is_super:
            raise AuthFailed(msg='只有超级管理员可操作')
        return fn(*args, **kwargs)

    return wrapper


def group_required(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        verify_jwt_in_request()
        current_user = get_current_user()
        # check current user is active or not
        # 判断当前用户是否为激活状态
        if not current_user.is_active:
            raise AuthFailed(msg='您目前处于未激活状态，请联系超级管理员')
        # not super
        if not current_user.is_super:
            group_id = current_user.group_id
            if group_id is None:
                raise AuthFailed(msg='您还不属于任何权限组，请联系超级管理员获得权限')
            from .core import is_user_allowed
            it = is_user_allowed(group_id)
            if not it:
                raise AuthFailed(msg='权限不够，请联系超级管理员获得权限')
            else:
                return fn(*args, **kwargs)
        else:
            return fn(*args, **kwargs)

    return wrapper


def login_required(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        verify_jwt_in_request()
        return fn(*args, **kwargs)

    return wrapper


@jwt.user_loader_callback_loader
def user_loader_callback(identity):
    from .core import find_user
    # token is granted , user must be exit
    # 如果token已经被颁发，则该用户一定存在
    user = find_user(id=identity)
    if user is None:
        raise NotFound(msg='用户不存在')
    return user


@jwt.expired_token_loader
def expired_loader_callback():
    return ExpiredTokenException()


@jwt.invalid_token_loader
def invalid_loader_callback(e):
    return InvalidTokenException()


@jwt.unauthorized_loader
def unauthorized_loader_callback(e):
    return AuthFailed(msg='认证失败，请检查请求头或者重新登陆')


def get_tokens(user):
    access_token = create_access_token(identity=user.id)
    refresh_token = create_refresh_token(identity=user.id)
    return access_token, refresh_token


# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

from flask_sqlalchemy import SQLAlchemy as _SQLAlchemy, BaseQuery
from sqlalchemy import inspect, orm, func
from contextlib import contextmanager

from .exception import NotFound


class SQLAlchemy(_SQLAlchemy):
    @contextmanager
    def auto_commit(self):
        try:
            yield
            self.session.commit()
        except Exception as e:
            self.session.rollback()
            raise e


class Query(BaseQuery):

    def filter_by(self, soft=True, **kwargs):
        # soft 应用软删除
        soft = kwargs.get('soft')
        if soft:
            kwargs['delete_time'] = None
        if soft is not None:
            kwargs.pop('soft')
        return super(Query, self).filter_by(**kwargs)

    def get_or_404(self, ident):
        rv = self.get(ident)
        if not rv:
            raise NotFound()
        return rv

    def first_or_404(self):
        rv = self.first()
        if not rv:
            raise NotFound()
        return rv


db = SQLAlchemy(query_class=Query)


def get_total_nums(cls, is_soft=False, **kwargs):
    nums = db.session.query(func.count(cls.id))
    nums = nums.filter(cls.delete_time == None).filter_by(**kwargs).scalar() if is_soft else nums.filter().scalar()
    if nums:
        return nums
    else:
        return 0


class MixinJSONSerializer:
    @orm.reconstructor
    def init_on_load(self):
        self._fields = []
        self._exclude = []

        self._set_fields()
        self.__prune_fields()

    def _set_fields(self):
        pass

    def __prune_fields(self):
        columns = inspect(self.__class__).columns
        if not self._fields:
            all_columns = set([column.name for column in columns])
            self._fields = list(all_columns - set(self._exclude))

    def hide(self, *args):
        for key in args:
            self._fields.remove(key)
        return self

    def keys(self):
        return self._fields

    def __getitem__(self, key):
        return getattr(self, key)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

import json
from collections import deque


class Sse(object):
    messages = deque()
    _retry = None

    def __init__(self, default_retry=2000):
        self._buffer = []
        self._default_id = 1
        self.set_retry(default_retry)

    def set_retry(self, num):
        self._retry = num
        self._buffer.append("retry: {0}\n".format(self._retry))

    def set_event_id(self, event_id=None):
        if event_id:
            self._default_id = event_id
            self._buffer.append("id: {0}\n".format(event_id))
        else:
            self._buffer.append("id: {0}\n".format(self._default_id))

    def reset_event_id(self):
        self.set_event_id(1)

    def increase_id(self):
        self._default_id += 1

    def add_message(self, event, obj, flush=True):
        self.set_event_id()
        self._buffer.append("event: {0}\n".format(event))
        line = json.dumps(obj, ensure_ascii=False)
        self._buffer.append("data: {0}\n".format(line))
        self._buffer.append("\n")
        if flush:
            self.flush()

    def flush(self):
        self.messages.append(self.join_buffer())
        self._buffer.clear()
        self.increase_id()

    def pop(self):
        return self.messages.popleft()

    def heartbeat(self, comment=None):
        # 发送注释 : this is a test stream\n\n 告诉客户端，服务器还活着
        if comment and type(comment) == 'str':
            self._buffer.append(comment)
        else:
            self._buffer.append(': sse sever is still alive \n\n')
        tmp = self.join_buffer()
        self._buffer.clear()
        return tmp

    def join_buffer(self):
        string = ''
        for it in self._buffer:
            string += it
        return string

    def exit_message(self):
        return len(self.messages) > 0


sser = Sse()



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
from datetime import datetime

from sqlalchemy import Column, Integer, String, FetchedValue, SmallInteger, TIMESTAMP
from werkzeug.security import generate_password_hash, check_password_hash

from .enums import UserSuper, UserActive
from .db import MixinJSONSerializer, db
from .util import camel2line


# 基础的crud model
class BaseCrud(db.Model, MixinJSONSerializer):
    __abstract__ = True

    def __init__(self):
        name: str = self.__class__.__name__
        if not hasattr(self, '__tablename__'):
            self.__tablename__ = camel2line(name)

    def _set_fields(self):
        self._exclude = []

    def set_attrs(self, attrs_dict):
        for key, value in attrs_dict.items():
            if hasattr(self, key) and key != 'id':
                setattr(self, key, value)

    # 硬删除
    def delete(self, commit=False):
        db.session.delete(self)
        if commit:
            db.session.commit()

    # 查
    @classmethod
    def get(cls, start=None, count=None, one=True, **kwargs):
        if one:
            return cls.query.filter().filter_by(**kwargs).first()
        return cls.query.filter().filter_by(**kwargs).offset(start).limit(count).all()

    # 增
    @classmethod
    def create(cls, **kwargs):
        one = cls()
        for key in kwargs.keys():
            if hasattr(one, key):
                setattr(one, key, kwargs[key])
        db.session.add(one)
        if kwargs.get('commit') is True:
            db.session.commit()
        return one

    def update(self, **kwargs):
        for key in kwargs.keys():
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
        db.session.add(self)
        if kwargs.get('commit') is True:
            db.session.commit()
        return self


# 提供软删除，及创建时间，更新时间信息的crud model
class InfoCrud(db.Model, MixinJSONSerializer):
    __abstract__ = True
    _create_time = Column('create_time', TIMESTAMP(True), default=datetime.now)
    update_time = Column(TIMESTAMP(True), default=datetime.now, onupdate=datetime.now)
    delete_time = Column(TIMESTAMP(True))

    def __init__(self):
        name: str = self.__class__.__name__
        if not hasattr(self, '__tablename__'):
            self.__tablename__ = camel2line(name)

    def _set_fields(self):
        self._exclude = ['update_time', 'delete_time']

    @property
    def create_time(self):
        if self._create_time is None:
            return None
        return int(round(self._create_time.timestamp() * 1000))

    # @property
    # def update_time(self):
    #     if self._update_time is None:
    #         return None
    #     return int(round(self._update_time.timestamp() * 1000))

    # @property
    # def delete_time(self):
    #     if self._delete_time is None:
    #         return None
    #     return int(round(self._delete_time.timestamp() * 1000))

    def set_attrs(self, attrs_dict):
        for key, value in attrs_dict.items():
            if hasattr(self, key) and key != 'id':
                setattr(self, key, value)

    # 软删除
    def delete(self, commit=False):
        self.delete_time = datetime.now()
        db.session.add(self)
        # 记得提交会话
        if commit:
            db.session.commit()

    # 硬删除
    def hard_delete(self, commit=False):
        db.session.delete(self)
        if commit:
            db.session.commit()

    # 查
    @classmethod
    def get(cls, start=None, count=None, one=True, **kwargs):
        # 应用软删除，必须带有delete_time
        if kwargs.get('delete_time') is None:
            kwargs['delete_time'] = None
        if one:
            return cls.query.filter().filter_by(**kwargs).first()
        return cls.query.filter().filter_by(**kwargs).offset(start).limit(count).all()

    # 增
    @classmethod
    def create(cls, **kwargs):
        one = cls()
        for key in kwargs.keys():
            # if key == 'from':
            #     setattr(one, '_from', kwargs[key])
            # if key == 'parts':
            #     setattr(one, '_parts', kwargs[key])
            if hasattr(one, key):
                setattr(one, key, kwargs[key])
        db.session.add(one)
        if kwargs.get('commit') is True:
            db.session.commit()
        return one

    def update(self, **kwargs):
        for key in kwargs.keys():
            # if key == 'from':
            #     setattr(self, '_from', kwargs[key])
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
        db.session.add(self)
        if kwargs.get('commit') is True:
            db.session.commit()
        return self


class UserInterface(InfoCrud):
    __tablename__ = 'jian_user'

    id = Column(Integer, primary_key=True)
    nickname = Column(String(24), nullable=False, unique=True)
    # : super express the user is super(super admin) ;  1 -> common |  2 -> super
    # : super 代表是否为超级管理员 ;  1 -> 普通用户 |  2 -> 超级管理员
    super = Column(SmallInteger, nullable=False, default=1, server_default=FetchedValue())
    # : active express the user can manage the authorities or not ; 1 -> active | 2 -> not
    # : active 代表当前用户是否为激活状态，非激活状态默认失去用户权限 ; 1 -> 激活 | 2 -> 非激活
    active = Column(SmallInteger, nullable=False, default=1, server_default=FetchedValue())
    # : used to send email in the future
    # : 预留字段，方便以后扩展
    email = Column(String(100), unique=True)
    # : which group the user belongs,nullable is true
    # : 用户所属的权限组id
    group_id = Column(Integer)
    _password = Column('password', String(100))

    def _set_fields(self):
        self._exclude = ['password']

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, raw):
        self._password = generate_password_hash(raw)

    @property
    def is_super(self):
        return self.super == UserSuper.SUPER.value

    @property
    def is_active(self):
        return self.active == UserActive.ACTIVE.value

    @classmethod
    def verify(cls, nickname, password):
        raise Exception('must implement this method')

    def check_password(self, raw):
        if not self._password:
            return False
        return check_password_hash(self._password, raw)

    def reset_password(self, new_password):
        raise Exception('must implement this method')

    def change_password(self, old_password, new_password):
        raise Exception('must implement this method')


class AuthInterface(BaseCrud):
    __tablename__ = 'jian_auth'

    id = Column(Integer, primary_key=True)
    # : belongs to which group
    # : 所属权限组id
    group_id = Column(Integer, nullable=False)
    # : authority field
    # : 权限字段
    auth = Column(String(60))
    # : authority module, default common , which can sort authorities
    # : 权限的模块
    module = Column(String(50))


class GroupInterface(BaseCrud):
    __tablename__ = 'jian_group'

    id = Column(Integer, primary_key=True)
    # : name of group
    # : 权限组名称
    name = Column(String(60))
    # a description of a group
    # 权限组描述
    info = Column(String(255))


class LogInterface(BaseCrud):
    __tablename__ = 'jian_log'

    id = Column(Integer, primary_key=True)
    # : log message
    # : 日志信息
    message = Column(String(450))
    # : create time
    # : 日志创建时间
    _time = Column('time', TIMESTAMP(True), default=datetime.now)
    # : user id
    # : 用户id
    user_id = Column(Integer, nullable=False)
    # user_name at that moment
    # 用户当时的昵称
    user_name = Column(String(20))
    # : status_code check request is success or not
    # : 请求的http返回码
    status_code = Column(Integer)
    # request method
    # 请求方法
    method = Column(String(20))
    # request path
    # 请求路径
    path = Column(String(50))
    # which authority is accessed
    # 访问那个权限
    authority = Column(String(100))

    @property
    def time(self):
        if self._time is None:
            return None
        return int(round(self._time.timestamp() * 1000))


class EventInterface(BaseCrud):
    __tablename__ = 'jian_event'
    id = Column(Integer, primary_key=True)
    # : belongs to which group
    group_id = Column(Integer, nullable=False)
    # message type ['订单','修改密码']
    message_events = Column(String(250))


# service暂时不用
class ServiceInterface(object):
    pass



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""
from .core import Jian, route_meta, manager
from .db import db
from .jwt import login_required, group_required, admin_required



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/07/03 <https://7yue.in>
"""

from flask import request
from wtforms import Form as WTForm, IntegerField
from wtforms.validators import StopValidation

from .exception import ParameterException


class Form(WTForm):
    def __init__(self):
        data = request.get_json(silent=True)
        args = request.args.to_dict()
        super(Form, self).__init__(data=data, **args)

    def validate_for_api(self):
        valid = super(Form, self).validate()
        if not valid:
            raise ParameterException(msg=self.errors)
        return self


def integer_check(form, field):
    if field.data is None:
        raise StopValidation('输入字段不可为空')
    try:
        field.data = int(field.data)
    except ValueError:
        raise StopValidation('不是一个有效整数')


class JianIntegerField(IntegerField):
    """
    校验一个字段是否为正整数
    """

    def __init__(self, label=None, validators=None, **kwargs):
        if validators is not None and type(validators) == list:
            validators.insert(0, integer_check)
        else:
            validators = [integer_check]
        super(JianIntegerField, self).__init__(label, validators, **kwargs)


# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/07/01 <https://7yue.in>
"""
from __future__ import print_function
from setuptools import setup

setup(
    name="Jian",
    version="0.1",
    author="Jarrott Xu",
    author_email="xqiqio7@gmail.com",
    description="Jarrott Flask中实现高可用的扩展",
    license="Apache License",
    url="",
    packages=['jian'],
    install_requires=[
        'requests',
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: Indexing",
        "Topic :: Utilities",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)


# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""


from functools import wraps
import re
from datetime import datetime
from flask import Response, request
from flask_jwt_extended import get_current_user
from .sse import sser

REG_XP = r'[{](.*?)[}]'
OBJECTS = ['user', 'response', 'request']
SUCCESS_STATUS = [200, 201]
MESSAGE_EVENTS = set()


class Notify(object):
    def __init__(self, template=None, event=None, **kwargs):
        """
        Notify a message or create a log
        :param template:  message template
        {user.nickname}查看自己是否为激活状态 ，状态码为{response.status_code} -> pedro查看自己是否为激活状态 ，状态码为200
        :param write: write to db or not
        :param push: push to front_end or not
        """
        if event:
            self.event = event
        elif self.event is None:
            raise Exception('event must not be None!')
        if template:
            self.template: str = template
        elif self.template is None:
            raise Exception('template must not be None!')
        # 加入所有types中
        MESSAGE_EVENTS.add(event)
        self.message = ''
        self.response = None
        self.user = None
        self.extra = kwargs

    def __call__(self, func):
        @wraps(func)
        def wrap(*args, **kwargs):
            response: Response = func(*args, **kwargs)
            self.response = response
            self.user = get_current_user()
            self.message = self._parse_template()
            self.push_message()
            return response

        return wrap

    def push_message(self):
        # status = '操作成功' if self.response.status_code in SUCCESS_STATUS else '操作失败'
        sser.add_message(self.event,
                         {'message': self.message,
                          'time': int(datetime.now().timestamp()),
                          **self.extra
                          })

    # 解析自定义消息的模板
    def _parse_template(self):
        message = self.template
        total = re.findall(REG_XP, message)
        for it in total:
            assert '.' in it, '%s中必须包含 . ,且为一个' % it
            i = it.rindex('.')
            obj = it[:i]
            assert obj in OBJECTS, '%s只能为user,response,request中的一个' % obj
            prop = it[i + 1:]
            if obj == 'user':
                item = getattr(self.user, prop, '')
            elif obj == 'response':
                item = getattr(self.response, prop, '')
            else:
                item = getattr(request, prop, '')
            message = message.replace('{%s}' % it, str(item))
        return message

    def _check_can_push(self):
        # 超级管理员不可push，暂时测试可push
        if self.user.is_super:
            return False
        return True



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""


class Plugin(object):
    def __init__(self, name=None):
        """
        :param name: plugin的名称
        """
        # container of plugin's controllers
        # 控制器容器
        self.controllers = {}
        # container of plugin's models
        # 模型层容器
        self.models = {}
        # plugin's services
        self.services = {}

        self.name = name

    def add_model(self, name, model):
        self.models[name] = model

    def get_model(self, name):
        return self.models.get(name)

    def add_controller(self, name, controller):
        self.controllers[name] = controller

    def add_service(self, name, service):
        self.services[name] = service

    def get_service(self, name):
        return self.services.get(name)



# -*- encoding:utf-8 -*-
"""
@ Created by Seven on  2019/01/19 <https://7yue.in>
"""

from functools import wraps
import re
from flask import Response, request
from flask_jwt_extended import get_current_user
from .core import find_info_by_ep, Log

REG_XP = r'[{](.*?)[}]'
OBJECTS = ['user', 'response', 'request']


class Logger(object):
    # message template
    template = None

    def __init__(self, template=None):
        if template:
            self.template: str = template
        elif self.template is None:
            raise Exception('template must not be None!')
        self.message = ''
        self.response = None
        self.user = None

    def __call__(self, func):
        @wraps(func)
        def wrap(*args, **kwargs):
            response: Response = func(*args, **kwargs)
            self.response = response
            self.user = get_current_user()
            self.message = self._parse_template()
            self.write_log()
            return response

        return wrap

    def write_log(self):
        info = find_info_by_ep(request.endpoint)
        authority = info.auth if info is not None else ''
        status_code = getattr(self.response, 'status_code', None)
        if status_code is None:
            status_code = getattr(self.response, 'code', None)
        if status_code is None:
            status_code = 0
        Log.create_log(message=self.message, user_id=self.user.id, user_name=self.user.nickname,
                       status_code=status_code, method=request.method,
                       path=request.path, authority=authority, commit=True)

    # 解析自定义模板
    def _parse_template(self):
        message = self.template
        total = re.findall(REG_XP, message)
        for it in total:
            assert '.' in it, '%s中必须包含 . ,且为一个' % it
            i = it.rindex('.')
            obj = it[:i]
            assert obj in OBJECTS, '%s只能为user,response,request中的一个' % obj
            prop = it[i + 1:]
            if obj == 'user':
                item = getattr(self.user, prop, '')
            elif obj == 'response':
                item = getattr(self.response, prop, '')
            else:
                item = getattr(request, prop, '')
            message = message.replace('{%s}' % it, str(item))
        return message
